@@ -1,41 +1,58 @@
 package RlwrapFilter;
 
+require 5.006;
+
 use strict;
 use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $AUTOLOAD);
 
 sub when_defined($@);
 my $previous_tag = -1;
-my $last_cumulative_output = "";
+my $echo_has_been_handled = 0;
+my $saved_output = "";
+
 
 require Exporter;
 require AutoLoader;
 @ISA = qw(Exporter AutoLoader);
 
-@EXPORT = qw(TAG_INPUT TAG_OUTPUT TAG_HISTORY TAG_COMPLETION TAG_PROMPT);
+@EXPORT = qw(TAG_INPUT TAG_OUTPUT TAG_HISTORY TAG_COMPLETION TAG_PROMPT TAG_HOTKEY TAG_SIGNAL DEBUG_FILTERING DEBUG_RANDOM_DELAY);
 $VERSION = '0.01';
 
 use Carp;
 
 # constants for every tag we know about
+use constant MAX_TAG                           => 255;
 use constant TAG_INPUT                         => 0;
 use constant TAG_OUTPUT                        => 1;
 use constant TAG_HISTORY                       => 2;
 use constant TAG_COMPLETION                    => 3;
 use constant TAG_PROMPT                        => 4;
+use constant TAG_HOTKEY                        => 5;
+use constant TAG_SIGNAL                        => 6;
+use constant TAG_WHAT_ARE_YOUR_INTERESTS       => 127;
 use constant TAG_IGNORE                        => 251;
 use constant TAG_ADD_TO_COMPLETION_LIST        => 252;
 use constant TAG_REMOVE_FROM_COMPLETION_LIST   => 253;
 use constant TAG_OUTPUT_OUT_OF_BAND            => 254;
 use constant TAG_ERROR                         => 255;
 
+# to bitwise AND with $ENV{RLWRAP_DEBUG}, the value of the rlwrap --debug option:
+use constant DEBUG_FILTERING                   => 16;
+use constant DEBUG_RANDOM_DELAY                => 1024;
+
+use constant REJECT_PROMPT                    => "_THIS_CANNOT_BE_A_PROMPT_";
+
+
 
 # we want to behave differently when running outside rlwrap
 my $we_are_running_under_rlwrap = defined $ENV{RLWRAP_COMMAND_PID};
 
 
-# die() and warn() must communicate via rlwrap, not via STDERR
-$SIG{__DIE__}  = \&die_with_error_message;
-$SIG{__WARN__} = \&warn_with_info_message;
+# die() and warn() must communicate via rlwrap, not via STDERR (unless we're running under perl -c)
+unless ($^C){
+  $SIG{__DIE__}  = \&die_with_error_message;
+  $SIG{__WARN__} = \&warn_with_info_message;
+}
 
 # automagically have a setter/getter for every key of %$self
 sub AUTOLOAD {
@@ -81,104 +98,156 @@ sub new {
   my $self = {};
   my @accessors = qw(initialiser help_text input_handler
 		   output_handler prompt_handler echo_handler
-		   message_handler history_handler completion_handler
+		   message_handler history_handler hotkey_handler completion_handler signal_handler
 		   echo_handler message_handler cloak_and_dagger_verbose
-		   cumulative_output prompts_are_never_empty);
+		   cumulative_output prompts_are_never_empty
+		   minimal_rlwrap_version);
   foreach my $acc (@accessors) {
     $self->{$acc} = "";
   }
+  bless $self, $class;
   foreach my $key (keys %init) {
     croak "There is no `$key' attribute in class $class" unless defined $self->{$key};
     $self -> {$key} = $init{$key};
+    $self -> minimal_rlwrap_version($self->{$key}) if $key eq "minimal_rlwrap_version";
   }
-  bless $self, $class;
+  return $self;
 }
 
-
-
 # event loop
+
 sub run {
-    my ($self) = @_;
+  my ($self) = @_;
 
-    if($ENV{RLWRAP_COMMAND_PID} == 0) { # when called as rlwrap -z <filter> (with no command) ..
-      write_message(TAG_OUTPUT_OUT_OF_BAND, $self -> help_text . "\n"); # ... send help text
+  if ($ENV{RLWRAP_COMMAND_PID} == 0) { # when called as rlwrap -z <filter> (with no command) ..
+    write_message(TAG_OUTPUT_OUT_OF_BAND, $self -> help_text . "\n"); # ... send help text
+  }
+
+  while (1) {
+     # make (un-parsed) $message accessible to filters
+    my ($tag, $message) = read_message();
+    $message = when_defined $self -> message_handler, "$message", $tag; # ignore return value
+    my $response = $message; # a response that is identical to the message signals: don't do anything
+
+    if ($tag == TAG_INPUT) {
+      $response = when_defined $self -> input_handler, "$message";
+    } elsif ($tag == TAG_OUTPUT) {
+      $response = $self -> handle_output($message);
+    } elsif ($tag == TAG_HISTORY) {
+      $response = when_defined $self -> history_handler, "$message";
+    } elsif ($tag == TAG_HOTKEY and $self -> hotkey_handler) {
+      my @params = split_rlwrap_message($message);
+      my @result = &{$self -> hotkey_handler}(@params);
+      $response = merge_fields(@result);
+    } elsif ($tag == TAG_COMPLETION and $self -> completion_handler) {
+      my ($line, $prefix, @completions) = split_rlwrap_message($message);
+      @completions = &{$self -> completion_handler}($line, $prefix, @completions);
+      $response = merge_fields($line, $prefix, @completions); # The handler only returns a (revised) list of completions. We add the (original) $line and $prefix
+    } elsif ($tag == TAG_PROMPT) {
+      if ($message eq REJECT_PROMPT or
+          ($self -> {prompts_are_never_empty} and $message eq "")) {
+        write_message($tag, REJECT_PROMPT);
+        # don't update <previous_tag> and don't reset <cumulative_input>
+        next;
+      }
+
+      if ($ENV{RLWRAP_IMPATIENT} and  $self->{cumulative_output} !~ /\n$/) { # cumulative output contains prompt: chop it off!
+        $self->{cumulative_output} =~ s/(?<![^\n])[^\n]*$// # s/[^\n]*$// takes way too long on big strings,
+          # what is the optimal regex to do this?
+      }
+
+      $response = when_defined $self -> prompt_handler, "$message";
+      croak "prompts may not contain newlines!" if $response =~ /\n/;
+    } elsif ($tag == TAG_SIGNAL) {
+      $response = when_defined($self -> signal_handler, $message);
+    } elsif ($tag == TAG_WHAT_ARE_YOUR_INTERESTS) {
+      $response = $self -> add_interests($message);
     }
 
-    while(1) {
-	my ($tag, $message) = read_message();
-	local $_ =  $message; # convenient when using anonymous subs as handlers: $filter -> blah_handler(sub{$_ if /blah/});
-        $message = when_defined $self -> message_handler, "$message", $tag; # ignore return value
-	my $response;
-	if ($tag == TAG_INPUT) {
-	  $response = when_defined $self -> input_handler, "$message";
-	} elsif ($tag == TAG_OUTPUT) {
-	  $response = $self -> handle_output($message);
-	} elsif ($tag == TAG_HISTORY) {
-	  $response = when_defined $self -> history_handler, "$message";
-	} elsif ($tag == TAG_COMPLETION) {
-	  my ($line, $prefix, $completions, @completions);
-	  if ($self -> completion_handler) {
-	    $message =~ s/[ ]+$//; # eat final space
-	    ($line, $prefix, $completions) = split /\t/, $message;
-	    @completions = split / /, $completions;
-	    @completions = &{$self -> completion_handler}($line, $prefix, @completions);
-	    $response = "$line\t$prefix\t". (join ' ', @completions) . " ";
-	  } else {
-	    $response = $message;
-	  }
-	} elsif ($tag == TAG_PROMPT) {
-	  if ($self -> {prompts_are_never_empty} and $message eq "") {
-			write_message($tag,"");
-			next;
-		}
-
-	  $self->{cumulative_output} =~ s/(?<![^\n])[^\n]*$// #  s/[^\n]*$// takes way too long on big strings, what is the optimal regex to do this?
-	    if $ENV{RLWRAP_IMPATIENT}; # chop off prompt from cumulative_output
-	  $response = when_defined $self -> prompt_handler, "$message";
-	  croak "prompts must not contain newlines!" if $response =~ /\n/;
-	} else {
-	  $response = $message; # No error message, compatible with future rlwrap
-	                        # versions that may define new tag types
-	}
-
-	out_of_band($tag) or $self -> {previous_tag} = $tag;
-	write_message($tag, $response);
 
+    # shouldn't the next "and" be an  "or"? @@@
+    unless (out_of_band($tag) and ($tag == TAG_PROMPT and $response eq REJECT_PROMPT)) {
+      $self -> {previous_tag} = $tag;
+      $self -> {previous_message} = $message;
     }
+    write_message($tag, $response);
+
+  }
 }
 
 
 # when_defined \&f, x, y, ... returns f(x, y, ...) if f is defined, x otherwise
 sub when_defined($@) {
   my $maybe_ref_to_sub = shift;
+  local $_ =  $_[0] ; # convenient when using anonymous subs as handlers: $filter -> blah_handler(sub{$_ if /blah/});
   if ($maybe_ref_to_sub) {
     if ((my $type  = ref($maybe_ref_to_sub))  ne 'CODE') {
       croak "improper handler <$maybe_ref_to_sub> of type $type (expected a ref to a sub)";
     }
     return &{$maybe_ref_to_sub}(@_);
   } else {
-    return shift;
+    return $_;
   }
 }
 
+# When the filter starts, it tells rlwrap its interests as a string 'yyny..' (1 + TAG_MAX chars, 1 for each tag)
+# when receiving a message 'nnynn...' the follwoing function changes 'n' to 'y' for those message types that the
+# filter handles,so that at the end of the pipeline the message reflects the interests of all filters in the
+# pipeline
+sub add_interests {
+  my ($self, $message) = @_;
+  my @interested = split //, $message;
+  for (my $tag = 0; $tag < @interested; $tag++) {
+    next if $interested[$tag] eq 'y'; # a preceding filter in the pipeline has already shown interest
+    $interested[$tag] = 'y'
+      if ($tag == TAG_INPUT      and $self -> input_handler)
+      or ($tag == TAG_OUTPUT     and ($self -> output_handler or $self -> echo_handler)) # echo is the first OUTPUT after INPUT
+      or ($tag == TAG_HISTORY    and ($self -> history_handler or $self -> echo_handler)) # to determine which OUTPUT is echo, we need to see INPUT
+      or ($tag == TAG_COMPLETION and $self -> completion_handler)
+      or ($tag == TAG_PROMPT     and $self -> prompt_handler)
+      or ($tag == TAG_HOTKEY     and $self -> hotkey_handler)
+      or ($tag == TAG_SIGNAL     and $self -> signal_handler);
+  }
+  return join '', @interested;
+}
+
+
 sub out_of_band {
   my($tag) = @_;
   return $tag > 128;
 }
 
+
+
+
+
+
 # split output in echo and the rest and call the appropriate handlers on them
 sub handle_output {
   my ($self, $message) = @_;
-  my ($echo, $handled_echo, $sep);
+  my ($echo, $handled_echo, $nl);
   if (defined $self -> {previous_tag} and $self -> {previous_tag} == TAG_INPUT) {
-      $self->{cumulative_output} = "";
-    ($echo, $sep, $message) = ($message =~ /^([^\n\r]*)(\r?\n)?(.*)?/s);
-    $handled_echo = when_defined $self -> echo_handler, "$echo";
+    $self->{cumulative_output} = "";
+    $echo_has_been_handled = 0;
+  }
+
+  if (not $echo_has_been_handled) {
+    if ($message !~ /\n/) {
+      $saved_output .= $message; # save all output until we have one *whole* echo line
+      return "";
+    } else {                    # ... then process it
+      $message = $saved_output . $message;
+      $echo_has_been_handled = 1;
+      $saved_output = "";
+      ($echo, $nl, $message) = ($message =~ /^([^\n\r]*)(\r?\n)?(.*)?/s); #@@@ This doesn't work for multi-line input!
+      $handled_echo = when_defined $self -> echo_handler, "$echo";
+    }
   }
   $self->{cumulative_output} .= $message;
-  return $handled_echo . $sep .(when_defined $self -> output_handler, "$message");
+  return $handled_echo . $nl . (when_defined $self -> output_handler, "$message");
 }
 
+
 sub read_until { # read chunks from pty pointed to by $fh until either inactive for $timeout or
                  # $stoptext is seen at end-of-chunk
   my ($fh, $stoptext, $timeout) = @_;
@@ -215,8 +284,9 @@ sub read_patiently {
   my $result;
   while($already_read < $count) {
     my $nread = sysread($fh, $result, $count-$already_read, $already_read);
-    if ($nread == 0) {
-      exit 0;
+    if ($nread == 0) { # rlwrap (or the rlwrapped command) has put down the
+                       # telephone - we're going to die anyway. Don't complain.
+      exit(0);
     } elsif ($nread < 0) {
       die_with_errormessage("error reading: $!");
     }
@@ -233,7 +303,7 @@ sub write_patiently {
   while($already_written < $count) {
     my $nwritten = syswrite($fh, $buffer, $count-$already_written, $already_written);
     if ($nwritten <= 0) {
-      die_with_errormessage("error writing: $!");
+      die_with_errormessage("error writing: $!\n");
     }
     $already_written += $nwritten;
   }
@@ -243,7 +313,6 @@ sub write_patiently {
 # read message (tag, length word and contents) from FILTER_IN
 sub read_message {
   return read_from_stdin() unless $we_are_running_under_rlwrap;
-
   my $tag = unpack("C", read_patiently(*FILTER_IN,1));
   my $length = unpack("L",read_patiently(*FILTER_IN,4));
   my $message = read_patiently(*FILTER_IN, $length);
@@ -269,6 +338,10 @@ sub read_from_stdin {
     print $prompt;
     ($tagname, $message) = (<STDIN> =~ /(\S+) (.*?)\r?\n/);
     exit unless $tagname;
+    $tagname = uc($tagname); # allow lowercase tag names 
+    $tagname = "TAG_" . $tagname unless $tagname =~ /^TAG_/; # allow omitting the tiresome "TAG_" prefix
+    $message =~ s/\\t/\t/g; # allow TABs to be input as '\t'
+    $message =~ s/\\n/\n/g; # the same for newlines
     $tag = name2tag(undef, $tagname); # call as function, not method
     $prompt = "again > ";
   }
@@ -306,7 +379,7 @@ sub cwd {
 # talking, and only listens, when it has displayed the $prompt
 sub cloak_and_dagger {
   my ($self, $question, $prompt, $timeout) = @_;
-  $prompt ||= $self -> last('prompt');
+  # $prompt ||= $self -> last('prompt'); TODO: remember last prompt
   write_patiently(*CMD_IN, "$question\n");
   $self -> send_output_oob("cloak_and_dagger question: $question\n") if $self -> {cloak_and_dagger_verbose};
   my $response = read_until(*CMD_OUT, $prompt, $timeout);
@@ -317,10 +390,26 @@ sub cloak_and_dagger {
 }
 
 
+# Commands return messages asynchronously and may time out
+# when invoked by multiple `cloak_and_dagger`. You may want to
+# drop their unused output at some later time:
+
+# rlwrap_filter.cloak_and_dagger($command1, $prompt, $timeout)
+# rlwrap_filter.cloak_and_dagger($command2, $prompt, $timeout) ...
+# sleep(1)
+# rlwrap_filter.vacuum_stale_message($prompt,  $timeout)
+
+sub vacuum_stale_message {
+  my ($self, $prompt, $timeout) = @_;
+  return read_until(*CMD_OUT, $prompt, $timeout);
+}
+
+
+
 sub tag2name {
   my ($self, $tag) = @_;
   for my $name (qw(TAG_REMOVE_FROM_COMPLETION_LIST TAG_ADD_TO_COMPLETION_LIST TAG_INPUT TAG_PROMPT TAG_COMPLETION
-		   TAG_HISTORY TAG_OUTPUT_OUT_OF_BAND TAG_ERROR  TAG_IGNORE TAG_OUTPUT)) {
+		   TAG_HOTKEY TAG_SIGNAL TAG_HISTORY TAG_WHAT_ARE_YOUR_INTERESTS  TAG_OUTPUT_OUT_OF_BAND TAG_ERROR  TAG_IGNORE TAG_OUTPUT)) {
     return $name if (eval "$tag == $name");
   }
   croak "unknown tag $tag";
@@ -338,29 +427,53 @@ sub send_output_oob {
   write_message(TAG_OUTPUT_OUT_OF_BAND, $text);
 }
 
-sub  send_ignore_oob {
+
+sub send_ignore_oob {
   my ($self, $text) = @_;
   write_message(TAG_IGNORE, $text);
 }
 
+sub tweak_readline_oob {
+  my ($self, $rl_tweak, @args) = @_;
+  my %nargs = (rl_variable_bind => 2,
+               rl_completer_word_break_characters => 1,
+               rl_completer_quote_characters => 1,
+               rl_filename_completion_desired => 1);
+               # the list can be extended in future versions
+
+  die "tweak_readline_oob() called with unknown readline function or variable '$rl_tweak'\n" unless  $nargs{$rl_tweak};
+  die "tweak_readline_oob($rl_tweak,...) should be called with $nargs{$rl_tweak} more args\n" if @args != $nargs{$rl_tweak};
+  $self -> send_ignore_oob("@" . join("::", ($rl_tweak, @args, "\n")));
+}
+
+
 sub die_with_error_message {
-	die $@ if $^S; # make die() within eval do the right thing
-  my ($mess) = @_;
+  my ($error_message) = @_;
+  die $error_message if $^S; # make die() within eval do the right thing
   my $myself = $0;
   $myself =~ s#^.*/([^.]+)$#$1#;
-  write_message(TAG_ERROR, "$myself: $mess");
+  write_message(TAG_ERROR, "$myself: $error_message");
+  sleep 2;
   exit 1;
 }
 
 
 sub warn_with_info_message {
-  my ($mess) = @_;
+  my ($warning) = @_;
   my $myself = $0;
   $myself =~ s#^.*/([^.]+)$#$1#;
-  write_message(TAG_OUTPUT_OUT_OF_BAND, "$myself: $mess");
+  write_message(TAG_OUTPUT_OUT_OF_BAND, "$myself: $warning");
 
 }
 
+sub minimal_rlwrap_version {
+  my ($self, $wanted) = @_;
+  my $found = $ENV{RLWRAP_VERSION} || "0.34";
+  die "This filter requires rlwrap version $wanted or newer!\n"
+    unless !$we_are_running_under_rlwrap or $wanted le $found;
+}
+
+
 sub command_line {
   my $commandline = $ENV{RLWRAP_COMMAND_LINE};
   return (wantarray ? split /\s+/, $commandline : $commandline);
@@ -370,12 +483,45 @@ sub running_under_rlwrap {
   return $we_are_running_under_rlwrap;
 }
 
+sub prompt_rejected {
+  my ($self) = @_;
+  $self->minimal_rlwrap_version("0.35");
+  return REJECT_PROMPT;
+}
+
 sub name {
   my ($name) = ($0 =~ m#([^/]+)$#);
   $name ||= $0;
   return $name;
 }
 
+use constant DIGIT_NUMBER => 8; # This has to be the same number as DIGITS_NUMBER  in src/string_utils.c:914
+
+
+sub split_rlwrap_message {
+  my ($message) = @_;
+  my @fields = ();
+
+  while(length($message) != 0){
+    my $lenstr = substr($message, 0, DIGIT_NUMBER, "");
+    my $len = hex($lenstr);
+    my $field = substr($message, 0, $len, "");
+    push(@fields, $field);
+  }
+  return @fields;
+}
+
+sub merge_fields {
+  my (@fields) = @_;
+  my $message = "";
+
+  foreach my $field (@fields) {
+    my $lenstr = sprintf("%0" . DIGIT_NUMBER . "x", length($field));
+    $message = $message . $lenstr . $field;
+  }
+  return $message;
+}
+
 
 1
 
@@ -401,7 +547,7 @@ RlwrapFilter - Perl class for B<rlwrap> filters
 
 =head1 DESCRIPTION
 
-B<rlwrap> (1) (L<http://utopia.knoware.nl/~hlub/uck/rlwrap>) is a tiny
+B<rlwrap> (1) (L<https://github.com/hanslub42/rlwrap>) is a tiny
 utility that sits between the user and any console command, in order
 to bestow readline capabilities (line editing, history recall) to
 commands that don't have them.
@@ -411,11 +557,17 @@ aspect of rlwrap's interaction with the user: changing the history,
 re-writing output and input, calling a pager or computing completion
 word lists from the current input.
 
+Filters can be combined in a pipeline using the special B<pipeline> filter.
+
 B<RlwrapFilter> makes it very simple to write rlwrap
 filters in perl. A filter only needs to instantiate a RlwrapFilter
 object, change a few of its default handlers and then call its 'run'
 method.
 
+There is also a Python 3 module B<rlwrapfilter.py>, distributed
+together with B<rlwrap>, that provides  more or less the same API as
+its B<perl> counterpart.
+
 =head1 PUBLIC METHODS
 
 =head2 CONSTRUCTOR
@@ -424,7 +576,7 @@ method.
 
 =item $f = new RlwrapFilter
 
-=item $f = RlwrapFilter -> new(prompt_handler => sub {"Hi! > "}, ...)
+=item $f = RlwrapFilter -> new(prompt_handler => sub {"Hi! > "}, minimal_rlwrap_version => "0.35", ...)
 
 Return a new RlwrapFilter object.
 
@@ -432,15 +584,34 @@ Return a new RlwrapFilter object.
 
 =head2 SETTING/GETTING HANDLERS
 
-Handlers are user-defined callbacks that get called from the
-'run' method with a message  (i.e. the un-filtered input,
-output, prompt) as their first argument. For convenience, $_ is set to the same
-value. They should return the re-written message text. They get called
-in a fixed cyclic order: prompt, completion, history, input, echo,
-output, prompt, ... etc ad infinitum. Rlwrap may always skip a handler when in direct mode,
-on the other hand, completion and output handlers may get called more
+Handlers are user-defined callbacks that specify one or more of an
+RlwrapFilter object's handler methods (handle_input, handle_prompt)
+They get called from the 'run' method in response to a message sent
+from B<rlwrap>.  Messages consist of a tag indicating which handler
+should be called (e.g. TAG_INPUT, TAG_HISTORY) and the message
+text. Usually, a filter overrides only one or at most two methods.
+
+=head2 CALLING CONVENTIONS
+
+In many cases (e.g. TAG_INPUT, TAG_OUTPUT, TAG_PROMPT) the message
+text is a simple string. Their handlers are called with the message
+text (i.e. the un-filtered input, output, prompt) as their only
+argument. For convenience, $_ is set to the same value. They should
+return the re-written message text.
+
+Some handlers (those for TAG_COMPLETION and TAG_HOTKEY) are a little
+more complex: their message text (accessible via $_) is a
+tab-separated list of fields; they get called with multiple arguments
+and are evaluated in list context.
+
+
+The message handlers are called in a fixed cyclic order: prompt,
+completion, history, input, echo, output, prompt, ... etc ad
+infinitum. Rlwrap may always skip a handler when in direct mode; on
+the other hand, completion and output handlers may get called more
 than once in succession. If a handler is left undefined, the result is
-as if the message text were returned unaltered.
+as if the message text were returned unaltered (in fact, B<rlwrap> knows
+when this is the case and won't even bother to send the message)
 
 It is important to note that the filter, and hence all its handlers,
 are bypassed when I<command> is in direct mode, i.e. when it asks for
@@ -466,21 +637,21 @@ the last output has arrived. Of course, B<rlwrap> cannot read the mind
 of I<command>, so what looks like a prompt to B<rlwrap> may actually
 be the beginning of an output line that took I<command> a little
 longer to formulate. If this is a problem, specify a longer "cooking"
-time with rlwrap's B<-w> option, or, if possible,  use the B<prompts_are_never_empty>
-method.
+time with rlwrap's B<-w> option, use the B<prompts_are_never_empty>
+method or "reject" the prompt (cf. the B<prompt_rejected> method)
 
 
 =item $handler = $f -> completion_handler, $f -> completion_handler(\&handler)
 
-The completion handler gets called with the the entire input line, the
-prefix (partial word to complete), and rlwrap's own completion list as
-arguments. It should return a (possibly revised) list of completions.  As
-an example, suppose the user has typed "She played for
+The completion handler gets called with three arguments: the entire input
+line, the prefix (partial word to complete), and rlwrap's own completion list.
+It should return a (possibly revised) list of completions.
+As an example, suppose the user has typed "She played for
 AE<lt>TABE<gt>". The handler will be called like this:
 
      myhandler("She played for A", "A", "Arsenal", "Arendal", "Anderlecht")
 
-it could then return a list of better clubs: ("Ajax", "AZ67",  "Arnhem")
+it could then return a list of stronger clubs: ("Ajax", "AZ67",  "Arnhem")
 
 =item $handler = $f -> history_handler, $f -> history_handler(\&handler)
 
@@ -488,10 +659,43 @@ Every input line is submitted to this handler, the return value is put
 in rlwrap's history. Returning an empty or undefined value will keep
 the input line out of the history.
 
+=item $handler = $f -> hotkey_handler, $f -> hotkey_handler(\&handler)
+
+If, while editing an input line, the user presses a key that is bound
+to "rlwrap_hotkey" in B<.inputrc>, the handler is called with five
+arguments: the hotkey, the prefix (i.e.  the part of the current input
+line before the cursor), the remaining part of the input line
+(postfix), the history as one string ("line 1\nline 2\n...line N", and
+the history position. It has to return a similar list, except that the
+first element will be printed in the "echo area" if it is changed from
+its original value.
+
+
+B<Example:> if the current input line is  "pea soup" (with the cursor on the
+space), and the user presses CTRL+P, which happens to be bound to "rlwrap-hotkey"
+in B<.inputrc>, the handler is called like this:
+
+    my_handler("\0x10", "pea", " soup", "tomato soup\nasparagus..", 12) # 16 = CTRL-P
+
+If you prefer peanut soup, the handler should return
+
+    ("Mmmm!", "peanut", " soup", "asparagus..", 11)
+
+after which the input line will be "peanut soup" (with the cursor
+again on the space), the echo area will display "Mmmm!", and any reference
+to inferior soups will have been purged from the history.
+
+If the returned input line ends with a newline B<rlwrap> will immediately
+accept the result.
+
+
+
 =item $handler = $f -> input_handler, $f -> input_handler(\&handler)
 
-Every input line is submitted to this handler, The handler's return
-value is written to I<command>'s pty (pseudo-terminal).
+Every input line (which may consist of multiple \n-separated lines,
+when using bracketed paste) is submitted to this handler, The
+handler's return value is written to I<command>'s pty
+(pseudo-terminal).
 
 =item $handler = $f -> echo_handler, $f -> echo_handler(\&handler)
 
@@ -501,6 +705,9 @@ it is the altered input that will be echo'ed back. If you don't want
 to confuse the user, use an echo handler that returns your original
 input.
 
+If you use rlwrap in --multi-line mode, additional echo lines will
+have to be handled by the output handler
+
 
 =item $handler = $f -> output_handler, $f -> output_handler(\&handler)
 
@@ -524,6 +731,20 @@ re-written output "out-of-band" just before the prompt:
                      });
 
 
+Note that when rlwrap is run in --multi-line mode the echo handler will still
+only handle the first echo line.  The remainder will generally
+be echoed back preceded by a continuation prompt; it is up to the
+output handler what to do with it.
+
+
+=item $handler = $f -> signal_handler, $f -> signal_handler(\&handler)
+
+As B<rlwrap> is transparent to signals, signals get passed on to I<command>.
+This handler gets called (as handler($signo)) for signals SIGHUP, SIGINT,
+SIGQUIT, SIGTERM, SIGCONT, SIGUSR1, SIGUSR2, and SIGWINCH, before the signal is delivered.
+It receives (and should return) $signo as a string. The returned signal is delivered to
+I<command>; return "0" to ignore the signal altogether. Output can be written out-of-band (to B<rlwrap>) or
+cloak_and_dagger (to I<command>, see below)
 
 
 =item $handler = $f -> message_handler, $f -> message_handler(\&handler)
@@ -534,8 +755,11 @@ all other handlers. Its return value is ignored, but it may be useful
 for logging and debugging purposes. The $tag is an integer that can be
 converted to a tag name by the 'tag2name' method
 
+
 =back
 
+
+
 =head2 OTHER METHODS
 
 =over 4
@@ -543,8 +767,12 @@ converted to a tag name by the 'tag2name' method
 =item $f -> help_text("Usage...")
 
 Set the help text for this filter. It will be displayed by rlwrap -z
-<filter>. The second line of the help text is used by C<rlwrap -z listing>,
-so please make it a short description of what the filter does.
+<filter>. The second line of the help text is used by C<rlwrap -z listing>;
+it should be a short description of what the filter does.
+
+=item $f -> minimal_rlwrap_version("x.yy")
+
+Die unless rlwrap is version x.yy or newer
 
 =item $dir = $f -> cwd
 
@@ -560,22 +788,13 @@ directory, but not its name)
 return the current cumulative output. All (untreated) output gets
 appended to the cumulative output after the output_handler has been
 called. The cumulative output starts with a fresh slate with every
-OUTPUT message that follows a non-OUTPUT message (ignoring out-of-band
-messages).
+OUTPUT message that directly follows an INPUT message (ignoring out-of-band
+messages and rejected prompts)
 
 When necessary (i.e. when B<rlwrap> is in "impatient mode") the prompt
 is removed from $filter->cumulative_output by the time the prompt
 handler is called.
 
-
-=item $text = $f -> prompts_are_never_empty($val)
-
-If $val evaluates to a true value, assume that I<command>s prompt is
-never empty. Then, whenever the filter receives an empty TAG_PROMPT
-message, it can safely assume that I<command> is in the middle of some
-output and only needs a bit more time. The prompt handler will not be
-called, and the cumulative output will not be reset. (default: 0)
-
 =item $tag = $f -> previous_tag
 
 The tag of the last preceding in-band message. A tag is an integer between 0 and
@@ -591,7 +810,7 @@ Convert a valid tag name like "TAG_PROMPT" to a tag (an integer)
 
 =item $f -> send_output_oob($text)
 
-Make rlwrap display C<$text>. C<$text> is sent "out-of-band":
+Make rlwrap display C<$text>. C<$text> is sent "out-of-band" :
 B<rlwrap> will not see it until just  after it has sent the next
 message to the filter
 
@@ -600,6 +819,24 @@ message to the filter
 Send an out-of-band TAG_IGNORE message to rlwrap. B<rlwrap> will silently
 discard it, but it can be useful when debugging filters
 
+
+=item $f -> tweak_readline_oob($readline_function, @parameters)
+
+Send a specially formatted out-of-band message in order to tweak
+readline (i.e. to make B<rlwrap> call a B<readline> function or set a
+B<readline> variable). See the GNU B<Readline> documentation for details.
+
+At this moment, the following tweaks are recognised:
+
+    $filter -> tweak_readline_oob("rl_variable_bind", $rl_variable_name, $value);
+    # ... only for bindable readline variables like those in .inputrc
+    $filter -> tweak_readline_oob("rl_completer_word_break_characters", $chars);
+    $filter -> tweak_readline_oob("rl_completer_quote_characters", $chars);
+    $filter -> tweak_readline_oob("rl_filename_completion_desired", "0" or "1");
+
+The parameters should not contain "::" (two consecutive colons). This method can be called at any
+moment, even before $filter -> run
+
 =item $f -> add_to_completion_list(@words)
 
 =item $f -> remove_from_completion_list(@words)
@@ -612,7 +849,8 @@ Send C<$question> to I<command>'s input and read back everything that
 comes back until C<$prompt> is seen at "end-of-chunk", or no new
 chunks arrive for $timeout seconds, whichever comes first.  Return the
 response (without the final C<$prompt>).  B<rlwrap> remains completely
-unaware of this conversation.
+unaware of this conversation; it won't even show up in B<rlwrap>'s debug logs
+or in the output of the B<logger> filter. Hence we also have:
 
 =item $f -> cloak_and_dagger_verbose($verbosity)
 
@@ -622,6 +860,16 @@ I<command>'s responses. By default, $verbosity = 0; setting it to
 1 will mess up the screen but greatly facilitate the (otherwise rather tricky) use of
 C<cloak_and_dagger>
 
+=item $self -> prompt_rejected
+
+A special text ("_THIS_CANNOT_BE_A_PROMPT_") to be returned by a
+prompt handler to "reject" the prompt. This will make rlwrap skip
+cooking the prompt.  $self->previous_tag and $self->cumulative_output
+will not be touched.
+
+=item $text = $f -> prompts_are_never_empty($val)
+
+If $val evaluates to a true value, automatically reject empty prompts.
 
 =item $f -> command_line
 
@@ -644,24 +892,25 @@ pipe.  This method never returns.
 
 =head1 LOW LEVEL PROTOCOL
 
-rlwrap communicates with a filter through messages consisting of a
-tag byte (TAG_OUTPUT, TAG_PROMPT etc. - to inform the filter of
-what is being sent), a unsigned 32-bit integer containing the message
-length, the message text and an extra newline. For every message sent, rlwrap expects,
-and waits for an answer message with the same tag. Sending back a
-different (in-band) tag is an error and instantly kills rlwrap, though filters
-may precede their answer message with "out-of-band" messages to output
-text (TAG_OUTPUT_OUT_OF_BAND), report errors (TAG_ERROR), and to
-manipulate the completion word list (TAG_ADD_TO_COMPLETION_LIST and
-TAG_REMOVE_FROM_COMPLETION_LIST) Out-of-band messages are not serviced
-by B<rlwrap> until right after it has sent the next in-band message - the
-communication with the filter is synchronous and driven by rlwrap.
+B<rlwrap> communicates with a filter through messages consisting of a tag
+byte (TAG_OUTPUT, TAG_PROMPT etc. - to inform the filter of what is
+being sent), an unsigned 32-bit integer containing the length of the
+message, the message text and an extra newline. For every message
+sent, rlwrap expects, and waits for an answer message with the same
+tag. Sending back a different (in-band) tag is an error and instantly
+kills rlwrap, though filters may precede their answer message with
+"out-of-band" messages to output text (TAG_OUTPUT_OUT_OF_BAND), report
+errors (TAG_ERROR), and to manipulate the completion word list
+(TAG_ADD_TO_COMPLETION_LIST and TAG_REMOVE_FROM_COMPLETION_LIST)
+Out-of-band messages are not serviced by B<rlwrap> until right after
+it has sent the next in-band message - the communication with the
+filter is synchronous and driven by rlwrap.
 
-Messages are received and sent via two pipes. The user's terminal is
-still connected to STDIN, STDOUT and STDERR, and you can read and
-write them directly, though this may mess up the screen and confuse
-the user unless you are careful. A filter can even communicate with
-the rlwrapped command behind rlwrap's back (cf the cloak_and_dagger()
+Messages are received and sent via two pipes. STDIN, STDOUT and STDERR
+are still connected to the user's terminal, and you can read and write
+them directly, though this may mess up the screen and confuse the user
+unless you are careful. A filter can even communicate with the
+rlwrapped command behind rlwrap's back (cf the cloak_and_dagger()
 method)
 
 The protocol uses the following tags (tags E<gt> 128 are out-of-band)
@@ -671,6 +920,10 @@ The protocol uses the following tags (tags E<gt> 128 are out-of-band)
  TAG_HISTORY     2
  TAG_COMPLETION  3
  TAG_PROMPT      4
+ TAG_HOTKEY      5
+ TAG_SIGNAL      6
+
+ TAG_WHAT_ARE_YOUR_INTERESTS     127
 
  TAG_IGNORE                      251
  TAG_ADD_TO_COMPLETION_LIST      252
@@ -680,10 +933,14 @@ The protocol uses the following tags (tags E<gt> 128 are out-of-band)
 
 
 To see how this works, you can eavesdrop on the protocol
-using the 'logger' filter.
+using the B<logger> filter.
 
 The constants TAG_INPUT, ... are exported by the RlwrapFilter.pm module.
 
+TAG_WHAT_ARE_YOUR_INTERESTS is only ever used internally, to prevent the exchange of messages that
+won't be handled by the filter anyway. It will be seen by the general message handler, and therefore show
+up (exactly once, at program start) in the output of e.g. the B<logger> filter.
+
 =head1 SIGNALS
 
 As STDIN is still connected to the users teminal, one might expect the filter
@@ -698,11 +955,12 @@ should unblock signals like SIGTERM, SIGWINCH.
 
 =head1 FILTER LIFETIME
 
-The filter is started by B<rlwrap> after I<command>. The cwd and cloak_and_dagger
-methods are immediately usable. When I<command> exits, the filter stays alive for
-a little longer in order to process I<command>'s last words. As calling the cwd and
-cloak_and_dagger methods at that time will make the filter die with an error,
-it may be advisable to wrap those calls in eval{}
+The filter is started by B<rlwrap> after I<command>, and stays alive
+as long as B<rlwrap> runs. Filter methods are immediately usable. When
+I<command> exits, the filter stays around for a little longer in order
+to process I<command>'s last words. As calling the cwd and
+cloak_and_dagger methods at that time will make the filter die with an
+error, it may be advisable to wrap those calls in eval{}
 
 If a filter calls die() it will send an (out-of-band) TAG_ERROR
 message to rlwrap before exiting. rlwrap will then report the message
@@ -715,16 +973,18 @@ die() within an eval() sets $@ as usual.
 
 Before calling a filter, B<rlwrap> sets the following environment variables:
 
-    RLWRAP_FILTERDIR      directory where RlwrapFilter.pm and most filters live (set by B<rlwrap>, can be
+    RLWRAP_FILTERDIR      directory where RlwrapFilter.pm and most filters live (set by rlwrap, can be
                           overridden by the user before calling rlwrap)
 
     PATH	          rlwrap automatically adds $RLWRAP_FILTERDIR to the front of filter's PATH
 
+    RLWRAP_VERSION        rlwrap version (e.g. "0.35")
+
     RLWRAP_COMMAND_PID    process ID of the rlwrapped command
 
     RLWRAP_COMMAND_LINE   command line of the rlwrapped command
 
-    RLWRAP_IMPATIENT      whether rlwrap is in "impatient mode" (cf B<rlwrap (1)>). In impatient mode,
+    RLWRAP_IMPATIENT      whether rlwrap is in "impatient mode" (cf rlwrap (1)). In impatient mode,
                           the candidate prompt is filtered through the output handler (and displayed before
                           being overwritten by the cooked prompt).
 
@@ -732,8 +992,11 @@ Before calling a filter, B<rlwrap> sets the following environment variables:
 
     RLWRAP_OUTPUT_PIPE_FD File descriptor of output pipe. For internal use only
 
-    RLWRAP_MASTER_PTY_FD File descriptor of I<command>'s pty.
+    RLWRAP_MASTER_PTY_FD  File descriptor of command's pty.
 
+    RLWRAP_BREAK_CHARS    The characters rlwrap considers word-breaking (cf. the --break-chars option in rlwrap (1))
+
+    RLWRAP_DEBUG          The value of the --debug (-d) option given to rlwrap
 
 =head1 DEBUGGING FILTERS
 
@@ -747,27 +1010,31 @@ the B<logger> filter, using a pipeline:
 
   rlwrap -z 'pipeline logger incoming : my_filter : logger outgoing' command
 
-Comparing the files C<incoming> and C<outgoing> will give you an exact idea of
-the filters behaviour.
 
 =head2 RUNNING WITHOUT B<rlwrap>
 
-When called by rlwrap filters get their input from the file descriptor
+When called by rlwrap, filters get their input from
 $RLWRAP_INPUT_PIPE_FD and write their output to
 $RLWRAP_OUTPUT_PIPE_FD, and expect and write messages consisting of a
 tag byte, a 32-bit length and the message proper. This is not terribly
 useful when running a filter directly from the command line (outside
 rlwrap), even if we set the RLWRAP_*_FD ourselves.
 
-Therfore, when run directly from the command line, a filter expects
+Therefore, when run directly from the command line, a filter expects
 input messages on its standard input of the form
 
-TAG_PROMPT >
+  TAG_PROMPT myprompt >
 
-(i.a. a tag name, one space and a message) and it will respond in the
-same way on its standard output
+(i.a. a tag name, one space and a message followed by a newline.  The
+message will not contain the final newline) and it will respond in the
+same way on its standard output. Of course, B<rlwrap> can help with the
+tedious typing of tag names (and, for convenience, you
+can omit the "TAG_" prefix and type the tag name in lowercase)
 
+  rlwrap -i -f TAGNAMES -p -S "filter > "  ./filter_to_be_debugged
 
+Because B<rlwrap> cannot put TABs and newlines in input lines, filters will
+convert '\t' and '\n' into TAB and newline when run directly from the command line.
 
 =head1 SEE ALSO
 
